@@ -89,6 +89,7 @@ extern "C"
 
 #include <arpa/inet.h>
 #include <netdb.h>
+#include <ctype.h>
 
 /* RS6000 has sys/select.h */
 #ifdef HAVE_SYS_SELECT_H
@@ -113,7 +114,6 @@ extern int h_errno;
 
 /*** Ping packet defines ***/
 
-#define MIN_PING_DATA   0
 #define MAX_IP_PACKET   65536   /* (theoretical) max IP packet size */
 #define SIZE_IP_HDR     20
 #ifndef IPV6
@@ -128,7 +128,7 @@ extern int h_errno;
 
 /* maxima and minima */
 #define MAX_COUNT               10000
-#define MIN_INTERVAL            10      /* in millisec */
+#define MIN_INTERVAL            1       /* in millisec */
 #define MIN_PERHOST_INTERVAL    20      /* in millisec */
 #define MIN_TIMEOUT             50      /* in millisec */
 #define MAX_RETRY               20
@@ -231,6 +231,7 @@ typedef struct host_entry
      int                  max_reply_i;        /* longest response time */
      int                  min_reply_i;        /* shortest response time */
      int                  total_time_i;       /* sum of response times */
+     int                  discard_next_recv_i;/* don't count next received reply for split reporting */
      int                  *resp_times;        /* individual response times */
 #if defined( DEBUG ) || defined( _DEBUG )
      int                  *sent_times;        /* per-sent-ping timestamp */
@@ -290,15 +291,16 @@ struct timeval current_time;        /* current time (pseudo) */
 struct timeval start_time; 
 struct timeval end_time;
 struct timeval last_send_time;      /* time last ping was sent */
-struct timeval last_report_time;    /* time last report was printed */
+struct timeval next_report_time;    /* time next -Q report is expected */
 struct timezone tz;
 
 /* switches */
 int generate_flag = 0;              /* flag for IP list generation */
 int verbose_flag, quiet_flag, stats_flag, unreachable_flag, alive_flag;
-int elapsed_flag, version_flag, count_flag, loop_flag;
+int elapsed_flag, version_flag, count_flag, loop_flag, netdata_flag;
 int per_recv_flag, report_all_rtts_flag, name_flag, addr_flag, backoff_flag;
 int multif_flag;
+int outage_flag = 0;
 int timestamp_flag = 0;
 int random_data_flag = 0;
 #if defined( DEBUG ) || defined( _DEBUG )
@@ -325,6 +327,7 @@ void usage( int );
 int wait_for_reply( long );
 void print_per_system_stats( void );
 void print_per_system_splits( void );
+void print_netdata( void );
 void print_global_stats( void );
 void main_loop();
 void finish();
@@ -362,14 +365,16 @@ int main( int argc, char **argv )
     int tos = 0; 
     HOST_ENTRY *cursor;
 
+    prog = argv[0];
+
     s = open_ping_socket(ping_data_size);
 
     if((uid = getuid())) {
         /* drop privileges */
-        setuid( getuid() );
+        if(setuid( getuid() ) == -1)
+            perror("cannot setuid");
     }
 
-    prog = argv[0];
     ident = getpid() & 0xFFFF;
     verbose_flag = 1;
     backoff_flag = 1;
@@ -377,10 +382,24 @@ int main( int argc, char **argv )
 
     /* get command line options */
 
-    while( ( c = getopt( argc, argv, "gedhlmnqusaAvDRz:t:H:i:p:f:r:c:b:C:Q:B:S:I:T:O:" ) ) != EOF )
+    while( ( c = getopt( argc, argv, "gedhlmnNqusaAvDRz:t:H:i:p:f:r:c:b:C:Q:B:S:I:T:O:M:o" ) ) != EOF )
     {
         switch( c )
         {
+        case 'M':
+#ifdef IP_MTU_DISCOVER
+            {
+                int val = IP_PMTUDISC_DO;
+                if (setsockopt(s, IPPROTO_IP, IP_MTU_DISCOVER, &val, sizeof(val))) {
+                    perror("setsockopt IP_MTU_DISCOVER");
+                }
+            }
+#else
+            fprintf(stderr, "-M option not supported on this platform\n");
+            exit(1);
+#endif
+            break;
+
         case 't':
             if( !( timeout = ( unsigned int )atoi( optarg ) * 100 ) )
                 usage(1);
@@ -388,14 +407,14 @@ int main( int argc, char **argv )
             break;
         
         case 'r':
-            if (!sscanf(optarg,"%i",&retry))
+            if (!sscanf(optarg,"%u",&retry))
                 usage(1);
             break;
         
         case 'i':
-            if( !( interval = ( unsigned int )atoi( optarg ) * 100 ) )
+            if (!sscanf(optarg,"%u",&interval))
                 usage(1);
-
+            interval *= 100;
             break;
 
         case 'p':
@@ -450,6 +469,10 @@ int main( int argc, char **argv )
             multif_flag = 1;
             break;
 
+        case 'N':
+            netdata_flag = 1;
+            break;
+
         case 'd': 
         case 'n':
             name_flag = 1;
@@ -553,6 +576,11 @@ int main( int argc, char **argv )
                 usage(1);
             }
             break;
+
+        case 'o':
+            outage_flag = 1;
+            break;
+
         default:
             fprintf(stderr, "see 'fping -h' for usage information\n");
             exit(1);
@@ -594,10 +622,10 @@ int main( int argc, char **argv )
         exit(1);
     }/* IF */
     
-    if( ( ping_data_size > MAX_PING_DATA ) || ( ping_data_size < MIN_PING_DATA ) )
+    if( ping_data_size > MAX_PING_DATA )
     {
-        fprintf( stderr, "%s: data size %u not valid, must be between %u and %u\n",
-            prog, ping_data_size, (unsigned int) MIN_PING_DATA, (unsigned int) MAX_PING_DATA );
+        fprintf( stderr, "%s: data size %u not valid, must be lower than %u\n",
+            prog, ping_data_size, (unsigned int) MAX_PING_DATA );
         exit(1);
     
     }/* IF */
@@ -695,6 +723,8 @@ int main( int argc, char **argv )
         if( randomly_lose_flag ) fprintf( stderr, "  randomly_lose_flag set\n" );
         if( sent_times_flag ) fprintf( stderr, "  sent_times_flag set\n" );
         if( print_per_system_flag ) fprintf( stderr, "  print_per_system_flag set\n" );
+        if( outage_flag ) fprintf( stderr, "  outage_flag set\n" );
+        if( netdata_flag ) fprintf( stderr, "  netdata_flag set\n" );
 
     }/* IF */
 #endif /* DEBUG || _DEBUG */
@@ -777,8 +807,9 @@ int main( int argc, char **argv )
         usage(1);
     }
     
-    if( !num_hosts )
-        exit(1);
+    if(!num_hosts) {
+        exit(num_noaddress ? 2 : 1);
+    }
 
     if(src_addr_present) {
         socket_set_src_addr(s, src_addr);
@@ -824,8 +855,10 @@ int main( int argc, char **argv )
     gettimeofday( &start_time, &tz );
     current_time = start_time;
 
-    if( report_interval )
-        last_report_time = start_time;
+    if( report_interval ) {
+        next_report_time = start_time;
+        timeval_add(&next_report_time, report_interval);
+    }
 
     last_send_time.tv_sec = current_time.tv_sec - 10000;
 
@@ -881,8 +914,8 @@ void add_cidr(char *addr)
     net_addr = ntohl(((struct sockaddr_in *) addr_res->ai_addr)->sin_addr.s_addr);
 
     /* check mask */
-    if(mask < 1 || mask > 30) {
-        fprintf(stderr, "Error: netmask must be between 1 and 30 (is: %s)\n", mask_str);
+    if(mask < 1 || mask > 32) {
+        fprintf(stderr, "Error: netmask must be between 1 and 32 (is: %s)\n", mask_str);
         exit(1);
     }
 
@@ -893,8 +926,14 @@ void add_cidr(char *addr)
     net_addr &= bitmask;
     net_last = net_addr + ((unsigned long) 0x1 << (32-mask)) - 1;
 
-    /* add all hosts in that network (excluding network and broadcast address) */
-    while(++net_addr < net_last) {
+    /* exclude network and broadcast address for regular prefixes */
+    if (mask < 31) {
+        net_last--;
+        net_addr++;
+    }
+
+    /* add all hosts in that network (net_addr and net_last inclusive) */
+    for(; net_addr <= net_last; net_addr++) {
         struct in_addr in_addr_tmp;
         char buffer[20];
         in_addr_tmp.s_addr = htonl(net_addr);
@@ -902,6 +941,7 @@ void add_cidr(char *addr)
         add_name(buffer);
     }
 
+
     freeaddrinfo(addr_res);
 }
 
@@ -959,6 +999,7 @@ void main_loop()
 {
     long lt;
     long wait_time;
+    long wait_time_next_report;
     HOST_ENTRY *h;
 
     while(ev_first) {
@@ -1063,6 +1104,15 @@ void main_loop()
             wait_time = interval;
         }
 
+        /* Make sure we don't wait too long, in case a report is expected */
+        if( report_interval && ( loop_flag || count_flag ) ) {
+               wait_time_next_report = timeval_diff ( &next_report_time, &current_time );
+               if(wait_time_next_report < wait_time) {
+                    wait_time = wait_time_next_report;
+                    if(wait_time < 0) { wait_time = 0; }
+               }
+        }
+
         /* Receive replies */
         /* (this is what sleeps during each loop iteration) */
         if(wait_for_reply(wait_time)) {
@@ -1073,10 +1123,15 @@ void main_loop()
 
         /* Print report */
         if( report_interval && ( loop_flag || count_flag ) &&
-            ( timeval_diff ( &current_time, &last_report_time ) > report_interval ) )
+            ( timeval_diff ( &current_time, &next_report_time ) >= 0 ) )
         {
-            print_per_system_splits();
-            last_report_time = current_time;
+            if(netdata_flag)
+                print_netdata();
+            else
+                print_per_system_splits();
+
+            while(timeval_diff ( &current_time, &next_report_time ) >= 0 )
+                timeval_add(&next_report_time, report_interval);
         }
     }
 }
@@ -1159,20 +1214,10 @@ void finish()
 
 void print_per_system_stats( void )
 {
-    int i, j, avg;
+    int i, j, avg, outage_ms;
     HOST_ENTRY *h;
-    char *buf;
-    int bufsize;
     int resp;
 
-    bufsize = max_hostname_len + 1;
-    buf = ( char* )malloc( bufsize );
-
-    if( !buf )
-        crash_and_burn( "can't malloc print buf" );
-
-    memset( buf, 0, bufsize );
-
     fflush( stdout );
 
     if( verbose_flag || per_recv_flag )
@@ -1183,49 +1228,47 @@ void print_per_system_stats( void )
         h = table[i];
         fprintf( stderr, "%s%s :", h->host, h->pad );
 
-        if( report_all_rtts_flag )
-        {
-            for( j = 0; j < h->num_sent; j++ )
-            {
+        if( report_all_rtts_flag ) {
+            for( j = 0; j < h->num_sent; j++ ) {
                 if( ( resp = h->resp_times[j] ) >= 0 )
                     fprintf( stderr, " %d.%02d", resp / 100, resp % 100 );
                 else
                     fprintf( stderr, " -" );
-
-            }/* FOR */
+            }
           
             fprintf( stderr, "\n" );
-      
-        }/* IF */
-        else
-        {
+        }
+        else {
             if( h->num_recv <= h->num_sent )
             {
                 fprintf( stderr, " xmt/rcv/%%loss = %d/%d/%d%%",
                     h->num_sent, h->num_recv, h->num_sent > 0 ?
                     ( ( h->num_sent - h->num_recv ) * 100 ) / h->num_sent : 0 );
 
-            }/* IF */
-            else
-            {
+                if (outage_flag) {
+                    /* Time outage total */
+                    outage_ms = (h->num_sent - h->num_recv) * perhost_interval/100;
+                    fprintf( stderr, ", outage(ms) = %d", outage_ms );
+                }
+
+            }
+            else {
                 fprintf( stderr, " xmt/rcv/%%return = %d/%d/%d%%",
                     h->num_sent, h->num_recv,
                     ( ( h->num_recv * 100 ) / h->num_sent ) );
           
-            }/* ELSE */
+            }
           
-            if( h->num_recv )
-            {
+            if( h->num_recv ) {
                 avg = h->total_time / h->num_recv;
                 fprintf( stderr, ", min/avg/max = %s", sprint_tm( h->min_reply ) );
                 fprintf( stderr, "/%s", sprint_tm( avg ) );
                 fprintf( stderr, "/%s", sprint_tm( h->max_reply ) );
           
-            }/* IF */
+            }
           
             fprintf(stderr, "\n");
-
-        }/* ELSE */
+        }
 
 #if defined( DEBUG ) || defined( _DEBUG )
         if( sent_times_flag )
@@ -1239,15 +1282,102 @@ void print_per_system_stats( void )
               
                 fprintf( stderr, "\n" );
 
-            }/* FOR */
+            }
+        }
+#endif
+    }
+}
+
+
+/************************************************************
+
+  Function: print_netdata
+
+*************************************************************
+
+  Inputs:  void (none)
+
+  Description:
+
+
+************************************************************/
+
+void print_netdata( void )
+{
+    static int sent_charts = 0;
+
+    int i, avg;
+    HOST_ENTRY *h;
+
+    fflush( stdout );
+
+    for( i = 0; i < num_hosts; i++ )
+    {
+        h = table[i];
+
+        if(!sent_charts) {
+            printf("CHART fping.%s_packets '' 'FPing Packets for host %s' packets '%s' fping.packets line 110020 %d\n", h->name, h->host, h->name, report_interval / 100000);
+            printf("DIMENSION xmt sent absolute 1 1\n");
+            printf("DIMENSION rcv received absolute 1 1\n");
+        }
+
+        printf("BEGIN fping.%s_packets\n", h->name);
+        printf("SET xmt = %d\n", h->num_sent_i);
+        printf("SET rcv = %d\n", h->num_recv_i);
+        printf("END\n");
+
+        if(!sent_charts) {
+            printf("CHART fping.%s_quality '' 'FPing Quality for host %s' percentage '%s' fping.quality area 110010 %d\n", h->name, h->host, h->name, report_interval / 100000);
+            printf("DIMENSION returned '' absolute 1 1\n");
+            /* printf("DIMENSION lost '' absolute 1 1\n"); */
+        }
+
+        /* if we just sent the probe and didn't receive a reply, we shouldn't count it */
+        h->discard_next_recv_i=0;
+        if( h->waiting && timeval_diff(&current_time, &h->last_send_time) < h->timeout) {
+            if(h->num_sent_i) {
+                h->num_sent_i--;
+                h->discard_next_recv_i=1;
+            }
+        }
+
+
+        printf("BEGIN fping.%s_quality\n", h->name);
+/*
+        if( h->num_recv_i <= h->num_sent_i )
+            printf("SET lost = %d\n", h->num_sent_i > 0 ? ( ( h->num_sent_i - h->num_recv_i ) * 100 ) / h->num_sent_i : 0 );
+        else
+            printf("SET lost = 0\n");
+*/
+
+        printf("SET returned = %d\n", h->num_sent_i > 0 ? ( ( h->num_recv_i * 100 ) / h->num_sent_i ) : 0 );
+        printf("END\n");
+
+        if(!sent_charts) {
+            printf("CHART fping.%s_latency '' 'FPing Latency for host %s' ms '%s' fping.latency area 110000 %d\n", h->name, h->host, h->name, report_interval / 100000);
+            printf("DIMENSION min minimum absolute 10 1000\n");
+            printf("DIMENSION max maximum absolute 10 1000\n");
+            printf("DIMENSION avg average absolute 10 1000\n");
+        }
+
+        printf("BEGIN fping.%s_latency\n", h->name);
+        if( h->num_recv_i )
+        {
+            avg = h->total_time_i / h->num_recv_i;
+            printf("SET min = %d\n", h->min_reply_i);
+            printf("SET avg = %d\n", avg);
+            printf("SET max = %d\n", h->max_reply_i);
+
         }/* IF */
-#endif /* DEBUG || _DEBUG */
+        printf("END\n");
+
+        h->num_sent_i = h->num_recv_i = h->max_reply_i =
+        h->min_reply_i = h->total_time_i = 0;
+
     }/* FOR */
-  
-    free( buf );
-
-} /* print_per_system_stats() */
 
+    sent_charts = 1;
+} /* print_per_system_splits() */
 
 /************************************************************
 
@@ -1264,65 +1394,63 @@ void print_per_system_stats( void )
 
 void print_per_system_splits( void )
 {
-    int i, avg;
+    int i, avg, outage_ms_i;
     HOST_ENTRY *h;
-    char *buf;
-    int bufsize;
     struct tm *curr_tm;
 
-    bufsize = max_hostname_len + 1;
-    buf = ( char* )malloc( bufsize );
-    if( !buf )
-        crash_and_burn( "can't malloc print buf" );
-
-    memset( buf, 0, bufsize );
-
     fflush( stdout );
 
     if( verbose_flag || per_recv_flag )
         fprintf( stderr, "\n" );
 
+    gettimeofday( &current_time, &tz );
     curr_tm = localtime( ( time_t* )&current_time.tv_sec );
     fprintf( stderr, "[%2.2d:%2.2d:%2.2d]\n", curr_tm->tm_hour,
         curr_tm->tm_min, curr_tm->tm_sec );
 
-    for( i = 0; i < num_hosts; i++ )
-    {
+    for( i = 0; i < num_hosts; i++ ) {
         h = table[i];
         fprintf( stderr, "%s%s :", h->host, h->pad );
 
-        if( h->num_recv_i <= h->num_sent_i )
-        {
+        /* if we just sent the probe and didn't receive a reply, we shouldn't count it */
+        h->discard_next_recv_i=0;
+        if( h->waiting && timeval_diff(&current_time, &h->last_send_time) < h->timeout) {
+            if(h->num_sent_i) {
+                h->num_sent_i--;
+                h->discard_next_recv_i=1;
+            }
+        }
+
+        if( h->num_recv_i <= h->num_sent_i ) {
             fprintf( stderr, " xmt/rcv/%%loss = %d/%d/%d%%",
                 h->num_sent_i, h->num_recv_i, h->num_sent_i > 0 ?
                 ( ( h->num_sent_i - h->num_recv_i ) * 100 ) / h->num_sent_i : 0 );
 
-        }/* IF */
-        else
-        {
+            if (outage_flag) {
+                /* Time outage  */
+                outage_ms_i = (h->num_sent_i - h->num_recv_i) * perhost_interval/100;
+                fprintf( stderr, ", outage(ms) = %d", outage_ms_i );
+            }
+        }
+        else {
             fprintf( stderr, " xmt/rcv/%%return = %d/%d/%d%%",
                 h->num_sent_i, h->num_recv_i, h->num_sent_i > 0 ?
                 ( ( h->num_recv_i * 100 ) / h->num_sent_i ) : 0 );
+        }
 
-        }/* ELSE */
-
-        if( h->num_recv_i )
-        {
+        if( h->num_recv_i ) {
             avg = h->total_time_i / h->num_recv_i;
             fprintf( stderr, ", min/avg/max = %s", sprint_tm( h->min_reply_i ) );
             fprintf( stderr, "/%s", sprint_tm( avg ) );
             fprintf( stderr, "/%s", sprint_tm( h->max_reply_i ) );
         
-        }/* IF */
+        }
         
         fprintf( stderr, "\n" );
         h->num_sent_i = h->num_recv_i = h->max_reply_i =
             h->min_reply_i = h->total_time_i = 0;
     
-    }/* FOR */
-
-    free( buf );
-
+    }
 } /* print_per_system_splits() */
 
 
@@ -1439,6 +1567,7 @@ int send_ping( int s, HOST_ENTRY *h )
     h->waiting++;
     num_pingsent++;
     last_send_time = h->last_send_time;
+    h->discard_next_recv_i=0;
 
     return(ret);
 }
@@ -1465,7 +1594,6 @@ int wait_for_reply(long wait_time)
     static char buffer[4096];
     struct sockaddr_storage response_addr;
     socklen_t response_addr_len;
-    struct ip *ip;
     int hlen = 0;
     FPING_ICMPHDR *icp;
     int n, avg;
@@ -1474,6 +1602,9 @@ int wait_for_reply(long wait_time)
     int this_count;
     struct timeval *sent_time;
     SEQMAP_VALUE *seqmap_value;
+#ifndef IPV6
+    struct ip *ip;
+#endif
 
     response_addr_len = sizeof(struct sockaddr_storage);
     result = recvfrom_wto( s, buffer, sizeof(buffer), (struct sockaddr *) &response_addr, &response_addr_len, wait_time );
@@ -1490,9 +1621,8 @@ int wait_for_reply(long wait_time)
     }
 #endif
 
-    ip = ( struct ip* )buffer;
-
 #ifndef IPV6
+    ip = ( struct ip* )buffer;
 #if defined( __alpha__ ) && __STDC__ && !defined( __GLIBC__ )
     /* The alpha headers are decidedly broken.
      * Using an ANSI compiler, it provides ip_vhl instead of ip_hl and
@@ -1561,17 +1691,23 @@ int wait_for_reply(long wait_time)
         h->waiting = 0;
         h->timeout = timeout;
         h->num_recv++;
-        h->num_recv_i++;
+
+        if(h->discard_next_recv_i) {
+            h->discard_next_recv_i=0;
+        }
+        else {
+            h->num_recv_i++;
+            if( !h->max_reply_i || this_reply > h->max_reply_i ) h->max_reply_i = this_reply;
+            if( !h->min_reply_i || this_reply < h->min_reply_i ) h->min_reply_i = this_reply;
+            h->total_time_i += this_reply;
+        }
 
         if( !max_reply      || this_reply > max_reply ) max_reply = this_reply;
         if( !min_reply      || this_reply < min_reply ) min_reply = this_reply;
         if( !h->max_reply   || this_reply > h->max_reply ) h->max_reply = this_reply;
         if( !h->min_reply   || this_reply < h->min_reply ) h->min_reply = this_reply;
-        if( !h->max_reply_i || this_reply > h->max_reply_i ) h->max_reply_i = this_reply;
-        if( !h->min_reply_i || this_reply < h->min_reply_i ) h->min_reply_i = this_reply;
         sum_replies += this_reply;
         h->total_time += this_reply;
-        h->total_time_i += this_reply;
         total_replies++;
     }
 
@@ -1818,7 +1954,6 @@ void add_name( char *name )
 {
     struct addrinfo   *res0, *res, hints;
     int               ret_ga;
-    size_t            len;
     char              *printname;
     char              namebuf[256];
     char              addrbuf[256];
@@ -1848,7 +1983,6 @@ void add_name( char *name )
     // (need to implement a separate option for this)
 
     for (res = res0; res; res = 0) {
-        len = res->ai_addrlen;
         /* name_flag: addr -> name lookup requested) */
         if(!name_flag) {
             printname = name;
@@ -1930,6 +2064,14 @@ void add_addr( char *name, char *host, struct sockaddr *ipaddr, socklen_t ipaddr
     p->running = 1;
     p->min_reply = 0;
 
+    if(netdata_flag) {
+        char *s = p->name;
+        while(*s) {
+            if(!isalnum(*s)) *s = '_';
+            s++;
+        }
+    }
+
     if( strlen( p->host ) > max_hostname_len )
         max_hostname_len = strlen( p->host );
 
@@ -2380,7 +2522,10 @@ void usage(int is_error)
 #endif
     fprintf(out, "   -l         loop sending pings forever\n" );
     fprintf(out, "   -m         ping multiple interfaces on target host\n" );
+    fprintf(out, "   -M         set the Don't Fragment flag\n" );
     fprintf(out, "   -n         show targets by name (-d is equivalent)\n" );
+    fprintf(out, "   -N         output compatible for netdata (-l -Q are required)\n" );
+    fprintf(out, "   -o         show the accumulated outage time (lost packets * packet interval)\n" );
     fprintf(out, "   -O n       set the type of service (tos) flag on the ICMP packets\n" );
     fprintf(out, "   -p n       interval between ping packets to one target (in millisec)\n" );
     fprintf(out, "                (in looping and counting modes, default %d)\n", perhost_interval / 100 );
